from django.db import models
from django.db.models.query import RawQuerySet
from django.forms import model_to_dict

from .models import Log, LogConfig

TYPES_FK = ['AutoField', 'BigAutoField', 'ForeignKey']
TYPES_INT = ['IntegerField', 'PositiveSmallIntegerField']
TYPES_DECIMAL = ['DecimalField']
TYPES_FLOAT = ['FloatField']
TYPES_STR = ['CharField', 'TextField']
TYPES_BOOL = ['BooleanField']
TYPES_DATETIME = ['DateTimeField']
TYPES_DATE = ['DateField']
TYPES_TIME = ['TimeField']


def internal_type_to_log_type(field: models.Field):
    internal_type = field.get_internal_type()
    if internal_type in TYPES_FK:
        return 'fk'
    if internal_type in TYPES_INT:
        return 'int'
    if internal_type in TYPES_FLOAT:
        return 'float'
    if internal_type in TYPES_DECIMAL:
        return 'decimal'
    if internal_type in TYPES_BOOL:
        return 'bool'
    if internal_type in TYPES_STR:
        return 'str'
    if internal_type in TYPES_DATE:
        return 'date'
    if internal_type in TYPES_DATETIME:
        return 'datetime'


def log_state_to_dict(logs: RawQuerySet) -> tuple[dict, dict]:
    state = dict()
    info = dict()
    if not logs:
        return state, info
    for log in logs:
        state.update({log.field: log.cast_value()})
        info.update({log.field: {'new_value_type': log.new_value_type}})
    return state, info


def log_value_to_instance_value(log: Log):
    if log.new_value_type == 'bool':
        return bool(log.new_value == 'True')

    if log.new_value == '':
        return None
    if log.new_value_type in ['fk', 'int']:
        return int(log.new_value)


def get_instance_state(instance: models.Model):
    state = model_to_dict(instance)
    cleaned_state = dict()
    log_config = LogConfig.objects.filter(
        model=f'{instance._meta.app_label}.{instance._meta.model_name}'
    ).prefetch_related('fields').first()
    if not log_config:
        return cleaned_state
    all_fields = list(map(lambda x: x.name, instance._meta.get_fields()))
    if log_config.exclude_fields:
        fields_to_log = list(f for f in all_fields if f not in log_config.fields.all().values_list('field_name', flat=True))
    else:
        fields_to_log = log_config.fields.all().values_list('field_name', flat=True)

    for f, v in state.items():
        if isinstance(v, (list, tuple, dict)):
            continue
        if f not in fields_to_log:
            continue
        cleaned_state.update({f: v})
    return cleaned_state



import logging

from django.views.generic import FormView, View
from django.urls import reverse_lazy, reverse
from django.contrib.auth import get_user_model
from django.template.loader import render_to_string
from django.shortcuts import render

from accrete.contrib.system_mail.forms import SystemMailCreateForm
from . forms import UserRegistrationForm, ConfirmRegistrationForm, ResendConfirmationMailForm
from . import config

_logger = logging.getLogger(__name__)
User = get_user_model()


def send_confirmation_mail(user, request):
    base = request.build_absolute_uri('/').strip('/')
    endpoint = reverse(
        'user_registration:confirmation', kwargs={'token': user.token.token}
    )
    confirmation_url = f"{base}{endpoint}"

    body = render_to_string(
        config.ACCRETE_USER_REGISTRATION_TEMPLATE_NAME,
        context={'confirmation_url': confirmation_url}
    )

    mail = SystemMailCreateForm({
        'from_name': config.ACCRETE_USER_REGISTRATION_MAIL_FROM_NAME,
        'to_addr': user.email,
        'subject': config.ACCRETE_USER_REGISTRATION_MAIL_SUBJECT,
        'body': body
    })

    if mail.errors:
        _logger.error(
            f"Validation Error while creating system mail for user registration\n"
            f"{dict(mail.errors)}"
        )
        return

    mail.save()


class RegistrationView(FormView):
    form_class = UserRegistrationForm
    template_name = 'user_registration/registration.html'
    success_url = reverse_lazy('user_registration:registration_mail_sent')

    def dispatch(self, request, *args, **kwargs):
        if not config.ACCRETE_USER_REGISTRATION_ALLOWED:
            return self.http_method_not_allowed(request, *args, **kwargs)
        return super().dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        user = form.save()
        send_confirmation_mail(user, self.request)
        return render(
            self.request,
            'user_registration/registration_confirmation_mail_sent.html',
            {'email': user.email}
        )


class ResendConfirmationMailView(FormView):
    form_class = ResendConfirmationMailForm
    template_name = 'user_registration/confirmation_resend.html'

    def form_valid(self, form):
        email = form.cleaned_data['email']
        user = User.objects.filter(email=email).first()

        if user and not user.is_active:
            send_confirmation_mail(user, self.request)

        return render(
            self.request,
            'user_registration/registration_confirmation_mail_sent.html',
            {'email': email, 'resent': True}
        )


class ConfirmRegistrationView(View):
    @staticmethod
    def get(request, *args, **kwargs):
        form = ConfirmRegistrationForm({'token': kwargs.get('token')})
        if form.is_valid():
            form.save()
            return render(request, 'user_registration/confirmation_succeeded.html', {})
        return render(
            request, 'user_registration/confirmation_failed.html', {
                'form': form
            }
        )



import json
import logging
import ast

from django.utils.deprecation import MiddlewareMixin
from django.http.response import HttpResponseRedirect, HttpResponsePermanentRedirect

from accrete.tenant import get_tenant

_logger = logging.getLogger(__name__)


class HtmxRedirectMiddleware(MiddlewareMixin):

    @staticmethod
    def process_response(request, response):
        is_htmx = request.headers.get('HX-Request', 'false') == 'true'
        if not is_htmx:
            return response
        is_redirect = isinstance(response, HttpResponseRedirect)
        is_permanent_redirect = isinstance(response, HttpResponsePermanentRedirect)
        if is_redirect or is_permanent_redirect:
            response.status_code = 200
            header = 'HX-Location' if is_redirect else 'HX-Redirect'
            response[header] = response['Location']

        tenant = get_tenant()
        trigger = {'setTenant': tenant and tenant.id or 0}
        res_trigger = response.get('HX-Trigger')
        try:
            res_trigger = res_trigger and ast.literal_eval(res_trigger) or None
        except Exception as e:
            _logger.warning(e)
            res_trigger = None
        if res_trigger:
            if isinstance(res_trigger, dict):
                trigger.update(res_trigger)
            else:
                _logger.warning(
                    'Response Header "HX-Trigger" '
                    'could not be evaluated to be a dict and is ignored!'
                )
        response['HX-Trigger'] = json.dumps(trigger)
        return response



import logging
import re
from datetime import datetime, date, timedelta

from django.utils.translation import gettext_lazy as _
from django import template
from django.db.models import Manager, DecimalField, IntegerField, FloatField, Model
from django.apps import apps
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

_logger = logging.getLogger(__name__)
register = template.Library()


@register.simple_tag(name='combine_templates')
def combine_templates(template_name, request=None):
    html = ''
    for app in apps.app_configs.keys():
        try:
            html += render_to_string(
                f'{app}/{template_name}',
                request=request
            )
        except template.TemplateDoesNotExist:
            continue
    return mark_safe(html)


@register.filter(name='related_obj_url')
def related_obj_url(param: object, value: str):
    attr = getattr(param, value, False)
    if attr and hasattr(attr, 'get_absolute_url'):
        return attr.get_absolute_url()
    return '#'


@register.filter(name='get_attr')
def get_attr_from_string(param: object, value: str):

    def return_save(attr):
        if isinstance(attr, Manager) and hasattr(attr, 'related_model'):
            return attr.related_model
        if attr is None:
            return None
        if isinstance(attr, str):
            return attr
        return attr

    try:
        attribute = getattr(param, value, False)
    except (AttributeError,):
        _logger.warning(f'Object {param} has no attribute {value}')
        attribute = None
    return return_save(attribute)


@register.filter(name='has_attr')
def has_attr(param: object, value:str):
    return hasattr(param, value)


@register.filter(name='get_item')
def get_item(param, value):
    return param[value]


@register.filter(name='get_model')
def get_model(instance):
    return instance._meta.model


@register.filter(name='verbose_field_name')
def verbose_field_name(param: object, value: str):
    if not param:
        return ''
    if isinstance(param, Model):
        name = param._meta.get_field(value).verbose_name
    else:
        name = getattr(param.model, value).field.verbose_name
    return name


@register.filter(name='table_display')
def table_display(param: object, value: str):
    field = getattr(param, value)
    if isinstance(field, Manager):
        return mark_safe(f"""
            <span>            
                <div class="tags">
                    {''.join(['<span class="tag">{}</span>'.format(o) for o in field.all()])}
                </div>
            </span>
        """)
    return get_attr_from_string(param, value)


@register.filter(name='table_alignment')
def table_alignment(param, value):
    if isinstance(param, str):
        return 'left'
    if isinstance(param, Model):
        field = param._meta.get_field(value)
    else:
        field = getattr(param.model, value).field
    if isinstance(field, (DecimalField, IntegerField, FloatField)):
        return 'right'
    return 'left'


@register.filter(name='message_class')
def message_class(message):
    if message.level == 25:
        return 'is-success'
    if message.level == 30:
        return 'is-warning'
    if message.level == 40:
        return 'is-danger'


@register.filter(name='timedelta_cast')
def timedelta_cast(td: timedelta, code: str) -> str | None:
    if not isinstance(td, timedelta):
        return None
    codes = ['days', 'hours', 'minutes', 'seconds', 'microseconds']
    if code not in codes:
        return None
    return str(getattr(td, code))


@register.filter(name='weekday')
def datetime_to_weekday(dt: datetime|date, default=None) -> str:
    if dt is None:
        return default
    mapping = {
        1: _('Mon'),
        2: _('Tue'),
        3: _('Wed'),
        4: _('Thu'),
        5: _('Fri'),
        6: _('Sat'),
        7: _('Sun')
    }
    return mapping[dt.isoweekday()]

@register.filter(name='xrefsave')
def x_ref_save(param: str):
    return re.sub(r'[^A-Za-z]+','', param)



# Generated by Django 4.2.6 on 2023-10-10 20:06

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SystemMail',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('from_name', models.CharField(max_length=255, verbose_name='From')),
                ('to_addr', models.TextField(help_text='E-Mail addresses, comma separated', verbose_name='To')),
                ('subject', models.CharField(default='', max_length=255, verbose_name='Subject')),
                ('body', models.TextField(default='', verbose_name='Mail Body')),
                ('sent', models.BooleanField(default=False, verbose_name='E-Mail sent')),
                ('error', models.TextField(blank=True, null=True, verbose_name='Errors')),
            ],
            options={
                'db_table': 'system_mail',
            },
        ),
    ]



import logging

from django.utils.deprecation import MiddlewareMixin
from accrete.tenant import set_tenant, set_member, get_tenant

from .models import Tenant, Member

_logger = logging.getLogger(__name__)


class TenantMiddleware(MiddlewareMixin):
    @staticmethod
    def get_tenant_id_from_request(request):
        tenant_id = (
            request.GET.get('tenant_id')
            or request.META.get('HTTP_X_TENANT_ID')
        )
        try:
            tenant_id = int(tenant_id)
            assert tenant_id > 0
        except (ValueError, TypeError, AssertionError):
            tenant_id = None
        return tenant_id

    def process_request(self, request):
        request.tenant = None
        request.member = None
        set_member(False)
        if request.path.startswith('/admin'):
            return
        if not request.user.is_authenticated:
            set_member(None)
            return
        tenant_id = self.get_tenant_id_from_request(request)
        tenant = Tenant.objects.none()
        memberships = Member.objects.filter(
            user=request.user, is_active=True, tenant__is_active=True
        ).prefetch_related('tenant', 'user').order_by('pk')
        if tenant_id:
            tenant = Tenant.objects.get(pk=tenant_id)
            memberships = memberships.filter(tenant=tenant)
        if memberships.count() == 1:
            request.member = memberships.first()
            request.tenant = request.member.tenant
            set_member(request.member)
            self.update_post_data(request)
            return
        if memberships.count() > 1:
            set_member(None)
            return
        if request.user.is_staff and tenant:
            set_member(None)
            set_tenant(tenant)
            request.tenant = tenant
            return
        set_member(None)
        return

    @staticmethod
    def update_post_data(request):
        if request.POST and not request.POST.get('tenant') and request.tenant:
            request.POST = request.POST.copy()
            request.POST['tenant'] = request.tenant.pk

    @staticmethod
    def process_response(request, response):
        tenant = get_tenant()
        response['X-TENANT-ID'] = tenant and tenant.id or 0
        return response



from django.db import models
from django.db.models.functions import Lower
from django.utils.translation import gettext_lazy as _
from accrete.fields import TranslatedCharField


class Country(models.Model):

    class Meta:
        verbose_name = _('Country')
        verbose_name_plural = _('Countries')
        db_table = 'accrete_country'
        ordering = ['-order_priority', Lower('name')]

    name = TranslatedCharField(
        verbose_name=_('Name')
    )

    iso_code_a2 = models.CharField(
        verbose_name=_('ISO Code Alpha 2'),
        max_length=2
    )

    iso_code_a3 = models.CharField(
        verbose_name=_('ISO Code Alpha 3'),
        max_length=3
    )

    vat_prefix = models.CharField(
        verbose_name=_('VAT ID Prefix'),
        max_length=3,
        null=True,
        blank=True
    )

    order_priority = models.IntegerField(
        verbose_name=_('Order Priority'),
        default=0
    )

    def __str__(self):
        return self.name



import logging
import json
import operator
from typing import Callable

from django.core import paginator
from django.db.models import Model, Q, QuerySet
from django.template.loader import render_to_string, TemplateDoesNotExist
from django.http import HttpResponseNotAllowed
from accrete.utils.models import get_related_model

_logger = logging.getLogger(__name__)

QUERYSTRING_KEY_MAP = {
    'querystring': 'q',
    'order': 'order',
    'paginate_by': 'paginate_by',
    'page': 'page'
}


def page_from_querystring(
    model: type[Model], query_dict: dict, key_map: dict = None,
    select_related: list = None, prefetch_related: list = None
) -> paginator.Page:

    key_map = key_map or QUERYSTRING_KEY_MAP
    queryset = filter_from_querystring(
        model, query_dict, key_map
    ).select_related(
        *select_related or []
    ).prefetch_related(
        *prefetch_related or []
    )
    return get_page(
        queryset,
        cast_param(query_dict, key_map.get('paginate_by'), int, 40),
        cast_param(query_dict, key_map.get('page'), int, 1)
    )


def get_page(queryset: QuerySet, paginate_by: int, page_number: int) -> paginator.Page:
    pages = paginator.Paginator(queryset, per_page=paginate_by)
    return pages.page(
        page_number <= pages.num_pages and page_number or pages.num_pages
    )


def filter_from_querystring(
        model: type[Model], query_dict: dict, key_map: dict = None
) -> QuerySet:

    key_map = key_map or QUERYSTRING_KEY_MAP
    querystring = query_dict.get(key_map['querystring'], '[]')
    order = query_dict.get(key_map['order']) or model._meta.ordering

    return model.objects.filter(
        parse_querystring(model, querystring)
    ).order_by(*order).distinct()


def parse_querystring(model: type[Model], query_string: str) -> Q:
    """
    param: query_string: JSON serializable string
    [{"term": value}, "&", [{t"erm": value}, "|", {"~term": value}]]
    Q(term=value) & (Q(term=value) | ~Q(term=value))
    """

    def get_expression(term: str, value) -> Q:
        invert = False
        if term.startswith('~'):
            invert = True
            term = term[1:]

        parts = term.split('__')
        if len(parts) == 1:
            expression = Q(**{term: value})
            return ~expression if invert else expression

        rel_path = '__'.join(parts[:-2])
        related_model, x = get_related_model(model, rel_path)
        expression = Q(**{term: value})
        return ~expression if invert else expression

    def parse_query_block(sub_item) -> Q:
        op = ops['&']
        parsed_query = Q()
        for item in sub_item:
            if isinstance(item, list):
                parsed_query = op(parsed_query, parse_query_block(item))
                op = ops['&']
            elif isinstance(item, dict):
                dict_query = Q()
                for term, value in item.items():
                    dict_query = ops['&'](dict_query, get_expression(term, value))
                parsed_query = op(parsed_query, dict_query)
            elif isinstance(item, str):
                if item not in '&|^':
                    raise ValueError(
                        f'Invalid operator in querystring: {item}.'
                        f'Operator must be one of &, |, ^'
                    )
                op = ops[item]
            else:
                raise ValueError(
                    f'Unsupported item in querystring: {item}.'
                    f'Item must be an instance of list, dict or str'
                )
        return parsed_query

    query_data = json.loads(query_string.strip('?&='))
    if isinstance(query_data, dict):
        query_data = [query_data]

    ops = {'&': operator.and_, '|': operator.or_, '^': operator.xor}
    query = parse_query_block(query_data)
    return query


def cast_param(params: dict, param: str, cast_to: Callable, default):
    if param not in params:
        return default
    try:
        return cast_to(params.get(param, default))
    except Exception as e:
        _logger.exception(e)
        return default


def method_not_allowed(method: str, allowed: list[str]) -> HttpResponseNotAllowed | None:
    if method not in allowed:
        return HttpResponseNotAllowed(allowed)


def render_templates(
    templates: list[str | tuple[str, dict]],
    context: dict = None,
    request=None,
    log_not_found: bool = False
) -> str:
    context = {} if context is None else context
    content = ''
    for template in templates:
        if isinstance(template, tuple):
            template_name, template_context = template
        else:
            template_name = template
            template_context = context
        try:
            content += render_to_string(template_name, template_context, request)
        except TemplateDoesNotExist as e:
            if log_not_found:
                _logger.warning(repr(e))
        except Exception as e:
            _logger.exception(repr(e))
            raise e
    return content



from uuid import uuid4
from django.forms import widgets
from django.shortcuts import resolve_url


class ModelSearchSelect(widgets.NumberInput):
    template_name = 'ui/widgets/model_search_select.html'
    option_template_name = 'ui/widgets/model_search_select_options.html'
    input_type = 'number'

    def __init__(
        self,
        search_url: str,
        search_parameter: str = 'search',
        limit: int | None = 5,
        hx_trigger_on_change: bool = False,
        choices=()
    ):
        super().__init__()
        self.search_url = search_url
        self.search_parameter = search_parameter
        self.limit = limit
        self.hx_trigger_on_change = hx_trigger_on_change
        self.choices = choices

    def get_context(self, name, value, attrs):
        qs = self.choices.queryset
        if self.limit:
            qs = qs[:self.limit]
        uuid = str(uuid4())

        context = {
            "widget": {
                "name": name,
                "is_hidden": self.is_hidden,
                "required": self.is_required,
                "value": self.format_value(value or ''),
                'value_display': self.value_display(value),
                "attrs": self.build_attrs(self.attrs, attrs),
                "template_name": self.template_name,
                'search_url': self.search_url,
                'search_parameter': self.search_parameter,
                'hx_trigger_on_change': self.hx_trigger_on_change,
                'uuid': uuid
            },
            'options': qs
        }
        return context

    def format_value(self, value):
        res = super().format_value(value)
        return res or ''

    def value_display(self, value):
        if value is None:
            return ''
        try:
            value = int(value)
        except ValueError:
            return ''
        return str(self.choices.queryset.get(pk=value))


class ModelSearchSelectMulti(widgets.SelectMultiple):
    template_name = 'ui/widgets/model_search_select_multi.html'
    option_template_name = 'ui/widgets/model_search_select_options.html'
    input_type = 'number'

    def __init__(
        self,
        search_url: str,
        search_parameter: str = 'search',
        limit: int | None = 5,
        hx_trigger_on_change: bool = False,
        choices=()
    ):
        super().__init__()
        self.search_url = search_url
        self.search_parameter = search_parameter
        self.limit = limit
        self.hx_trigger_on_change = hx_trigger_on_change
        self.choices = choices

    def get_context(self, name, value, attrs):
        qs = self.choices.queryset
        if self.limit:
            qs = qs[:self.limit]
        uuid = uuid4()

        context = {
            "widget": {
                "name": name,
                "is_hidden": self.is_hidden,
                "required": self.is_required,
                "value": self.format_value(value),
                "attrs": self.build_attrs(self.attrs, attrs),
                "template_name": self.template_name,
                'search_url': resolve_url(self.search_url),
                'search_parameter': self.search_parameter,
                'hx_trigger_on_change': self.hx_trigger_on_change,
                'uuid': uuid
            },
            'options': qs,
            'selected': self.choices.queryset.filter(pk__in=value) if value else qs.none()
        }
        return context
